@@ -7,6 +7,10 @@
 #include <X11/keysym.h>
 #include <X11/XF86keysym.h>
 
+#include <string.h>
+#include <stdlib.h>
+#include <stdio.h>
+
 #include <sys/select.h>
 #include <sys/time.h>
 #include <sys/types.h>
@@ -88,7 +92,6 @@ static int transition_dirty = 0;
 static int dirtysearch = 0;
 static char *password = "";
 static XColor xbgcolor;
-static XColor xshcolor;
 static int reqw = 0;
 static int reqh = 0;
 static char copylatin1[1024 * 16] = "";
@@ -180,7 +183,7 @@ char *wintextinput(pdfapp_t *app, char *inittext, int retry)
 	return NULL;
 }
 
-int winchoiceinput(pdfapp_t *app, int nopts, char *opts[], int *nvals, char *vals[])
+int winchoiceinput(pdfapp_t *app, int nopts, const char *opts[], int *nvals, const char *vals[])
 {
 	/* FIXME: temporary dummy implementation */
 	return 0;
@@ -222,12 +225,7 @@ static void winopen(void)
 	xbgcolor.green = 0x7000;
 	xbgcolor.blue = 0x7000;
 
-	xshcolor.red = 0x4000;
-	xshcolor.green = 0x4000;
-	xshcolor.blue = 0x4000;
-
 	XAllocColor(xdpy, DefaultColormap(xdpy, xscr), &xbgcolor);
-	XAllocColor(xdpy, DefaultColormap(xdpy, xscr), &xshcolor);
 
 	xwin = XCreateWindow(xdpy, DefaultRootWindow(xdpy),
 		10, 10, 200, 100, 0,
@@ -313,13 +311,38 @@ void winreplacefile(char *source, char *target)
 
 void wincopyfile(char *source, char *target)
 {
-	char *buf = malloc(strlen(source)+strlen(target)+5);
-	if (buf)
+	FILE *in, *out;
+	char buf[32 << 10];
+	int n;
+
+	in = fopen(source, "rb");
+	if (!in)
 	{
-		sprintf(buf, "cp %s %s", source, target);
-		system(buf);
-		free(buf);
+		winerror(&gapp, "cannot open source file for copying");
+		return;
 	}
+	out = fopen(target, "wb");
+	if (!out)
+	{
+		winerror(&gapp, "cannot open target file for copying");
+		fclose(in);
+		return;
+	}
+
+	for (;;)
+	{
+		n = fread(buf, 1, sizeof buf, in);
+		fwrite(buf, 1, n, out);
+		if (n < sizeof buf)
+		{
+			if (ferror(in))
+				winerror(&gapp, "cannot read data from source file");
+			break;
+		}
+	}
+
+	fclose(out);
+	fclose(in);
 }
 
 void cleanup(pdfapp_t *app)
@@ -344,10 +367,10 @@ void cleanup(pdfapp_t *app)
 	fz_drop_context(ctx);
 }
 
-static int winresolution()
+static int winresolution(void)
 {
-	return DisplayWidth(xdpy, xscr) * 25.4 /
-		DisplayWidthMM(xdpy, xscr) + 0.5;
+	return DisplayWidth(xdpy, xscr) * 25.4f /
+		DisplayWidthMM(xdpy, xscr) + 0.5f;
 }
 
 void wincursor(pdfapp_t *app, int curs)
@@ -498,19 +521,18 @@ static void winblit(pdfapp_t *app)
 		int x1 = gapp.panx + image_w;
 		int y1 = gapp.pany + image_h;
 
-		XSetForeground(xdpy, xgc, xbgcolor.pixel);
+		if (app->invert)
+			XSetForeground(xdpy, xgc, BlackPixel(xdpy, DefaultScreen(xdpy)));
+		else
+			XSetForeground(xdpy, xgc, xbgcolor.pixel);
 		fillrect(0, 0, x0, gapp.winh);
 		fillrect(x1, 0, gapp.winw - x1, gapp.winh);
 		fillrect(0, 0, gapp.winw, y0);
 		fillrect(0, y1, gapp.winw, gapp.winh - y1);
 
-		XSetForeground(xdpy, xgc, xshcolor.pixel);
-		fillrect(x0+2, y1, image_w, 2);
-		fillrect(x1, y0+2, 2, image_h);
-
 		if (gapp.iscopying || justcopied)
 		{
-			pdfapp_invert(&gapp, &gapp.selr);
+			pdfapp_invert(&gapp, gapp.selr);
 			justcopied = 1;
 		}
 
@@ -553,7 +575,7 @@ static void winblit(pdfapp_t *app)
 
 		if (gapp.iscopying || justcopied)
 		{
-			pdfapp_invert(&gapp, &gapp.selr);
+			pdfapp_invert(&gapp, gapp.selr);
 			justcopied = 1;
 		}
 	}
@@ -751,6 +773,16 @@ void winopenuri(pdfapp_t *app, char *buf)
 	waitpid(pid, NULL, 0);
 }
 
+int winquery(pdfapp_t *app, const char *query)
+{
+	return QUERY_NO;
+}
+
+int wingetcertpath(char *buf, int len)
+{
+	return 0;
+}
+
 static void onkey(int c, int modifiers)
 {
 	advance_scheduled = 0;
@@ -803,9 +835,9 @@ static void signal_handler(int signal)
 		reloading = 1;
 }
 
-static void usage(void)
+static void usage(const char *argv0)
 {
-	fprintf(stderr, "usage: mupdf [options] file.pdf [page]\n");
+	fprintf(stderr, "usage: %s [options] file.pdf [page]\n", argv0);
 	fprintf(stderr, "\t-p -\tpassword\n");
 	fprintf(stderr, "\t-r -\tresolution\n");
 	fprintf(stderr, "\t-A -\tset anti-aliasing quality in bits (0=off, 8=best)\n");
@@ -836,7 +868,6 @@ int main(int argc, char **argv)
 	struct timeval now;
 	struct timeval *timeout;
 	struct timeval tmo_advance_delay;
-	int bps = 0;
 
 	ctx = fz_new_context(NULL, NULL, FZ_STORE_DEFAULT);
 	if (!ctx)
@@ -847,16 +878,14 @@ int main(int argc, char **argv)
 
 	pdfapp_init(ctx, &gapp);
 
-	while ((c = fz_getopt(argc, argv, "Ip:r:A:C:W:H:S:U:Xb:")) != -1)
+	while ((c = fz_getopt(argc, argv, "Ip:r:A:C:W:H:S:U:X")) != -1)
 	{
 		switch (c)
 		{
 		case 'C':
 			c = strtol(fz_optarg, NULL, 16);
 			gapp.tint = 1;
-			gapp.tint_r = (c >> 16) & 255;
-			gapp.tint_g = (c >> 8) & 255;
-			gapp.tint_b = (c) & 255;
+			gapp.tint_white = c;
 			break;
 		case 'p': password = fz_optarg; break;
 		case 'r': resolution = atoi(fz_optarg); break;
@@ -867,13 +896,12 @@ int main(int argc, char **argv)
 		case 'S': gapp.layout_em = fz_atof(fz_optarg); break;
 		case 'U': gapp.layout_css = fz_optarg; break;
 		case 'X': gapp.layout_use_doc_css = 0; break;
-		case 'b': bps = (fz_optarg && *fz_optarg) ? fz_atoi(fz_optarg) : 4096; break;
-		default: usage();
+		default: usage(argv[0]);
 		}
 	}
 
 	if (argc - fz_optind == 0)
-		usage();
+		usage(argv[0]);
 
 	filename = argv[fz_optind++];
 
@@ -892,6 +920,7 @@ int main(int argc, char **argv)
 	gapp.transitions_enabled = 1;
 	gapp.scrw = DisplayWidth(xdpy, xscr);
 	gapp.scrh = DisplayHeight(xdpy, xscr);
+	gapp.default_resolution = resolution;
 	gapp.resolution = resolution;
 	gapp.pageno = pageno;
 
@@ -899,10 +928,7 @@ int main(int argc, char **argv)
 	tmo_at.tv_usec = 0;
 	timeout = NULL;
 
-	if (bps)
-		pdfapp_open_progressive(&gapp, filename, 0, bps);
-	else
-		pdfapp_open(&gapp, filename, 0);
+	pdfapp_open(&gapp, filename, 0);
 
 	FD_ZERO(&fds);
 
@@ -943,24 +969,30 @@ int main(int argc, char **argv)
 						break;
 
 					case XK_Up:
+					case XK_KP_Up:
 						len = 1; buf[0] = 'k';
 						break;
 					case XK_Down:
+					case XK_KP_Down:
 						len = 1; buf[0] = 'j';
 						break;
 
 					case XK_Left:
+					case XK_KP_Left:
 						len = 1; buf[0] = 'b';
 						break;
 					case XK_Right:
+					case XK_KP_Right:
 						len = 1; buf[0] = ' ';
 						break;
 
 					case XK_Page_Up:
+					case XK_KP_Page_Up:
 					case XF86XK_Back:
 						len = 1; buf[0] = ',';
 						break;
 					case XK_Page_Down:
+					case XK_KP_Page_Down:
 					case XF86XK_Forward:
 						len = 1; buf[0] = '.';
 						break;
