@@ -1,13 +1,28 @@
 #include "mupdf/fitz.h"
 
+#include <string.h>
+
 #define RLE_THRESHOLD 256
 
+/*
+	Take a reference to a glyph.
+
+	pix: The glyph to increment the reference for.
+
+	Returns pix.
+*/
 fz_glyph *
 fz_keep_glyph(fz_context *ctx, fz_glyph *glyph)
 {
 	return fz_keep_storable(ctx, &glyph->storable);
 }
 
+/*
+	Drop a reference and free a glyph.
+
+	Decrement the reference count for the glyph. When no
+	references remain the glyph will be freed.
+*/
 void
 fz_drop_glyph(fz_context *ctx, fz_glyph *glyph)
 {
@@ -22,32 +37,40 @@ fz_drop_glyph_imp(fz_context *ctx, fz_storable *glyph_)
 	fz_free(ctx, glyph);
 }
 
-fz_irect *
-fz_glyph_bbox(fz_context *ctx, fz_glyph *glyph, fz_irect *bbox)
+fz_irect
+fz_glyph_bbox(fz_context *ctx, fz_glyph *glyph)
 {
-	bbox->x0 = glyph->x;
-	bbox->y0 = glyph->y;
-	bbox->x1 = glyph->x + glyph->w;
-	bbox->y1 = glyph->y + glyph->h;
+	fz_irect bbox;
+	bbox.x0 = glyph->x;
+	bbox.y0 = glyph->y;
+	bbox.x1 = glyph->x + glyph->w;
+	bbox.y1 = glyph->y + glyph->h;
 	return bbox;
 }
 
-fz_irect *
-fz_glyph_bbox_no_ctx(fz_glyph *glyph, fz_irect *bbox)
+fz_irect
+fz_glyph_bbox_no_ctx(fz_glyph *glyph)
 {
-	bbox->x0 = glyph->x;
-	bbox->y0 = glyph->y;
-	bbox->x1 = glyph->x + glyph->w;
-	bbox->y1 = glyph->y + glyph->h;
+	fz_irect bbox;
+	bbox.x0 = glyph->x;
+	bbox.y0 = glyph->y;
+	bbox.x1 = glyph->x + glyph->w;
+	bbox.y1 = glyph->y + glyph->h;
 	return bbox;
 }
 
+/*
+	Return the width of the glyph in pixels.
+*/
 int
 fz_glyph_width(fz_context *ctx, fz_glyph *glyph)
 {
 	return glyph->w;
 }
 
+/*
+	Return the height of the glyph in pixels.
+*/
 int
 fz_glyph_height(fz_context *ctx, fz_glyph *glyph)
 {
@@ -55,6 +78,8 @@ fz_glyph_height(fz_context *ctx, fz_glyph *glyph)
 }
 
 #ifndef NDEBUG
+#include <stdio.h>
+
 void
 fz_dump_glyph(fz_glyph *glyph)
 {
@@ -118,6 +143,12 @@ fz_dump_glyph(fz_glyph *glyph)
 }
 #endif
 
+/*
+	Create a new glyph from a pixmap
+
+	Returns a pointer to the new glyph. Throws exception on failure to
+	allocate.
+*/
 fz_glyph *
 fz_new_glyph_from_pixmap(fz_context *ctx, fz_pixmap *pix)
 {
@@ -156,6 +187,20 @@ fz_new_glyph_from_pixmap(fz_context *ctx, fz_pixmap *pix)
 	return glyph;
 }
 
+/*
+	Create a new glyph from 8bpp data
+
+	x, y: X and Y position for the glyph
+
+	w, h: Width and Height for the glyph
+
+	sp: Source Pointer to data
+
+	span: Increment from line to line of data
+
+	Returns a pointer to the new glyph. Throws exception on failure to
+	allocate.
+*/
 fz_glyph *
 fz_new_glyph_from_8bpp_data(fz_context *ctx, int x, int y, int w, int h, unsigned char *sp, int span)
 {
@@ -304,6 +349,20 @@ try_pixmap:
 	return glyph;
 }
 
+/*
+	Create a new glyph from 1bpp data
+
+	x, y: X and Y position for the glyph
+
+	w, h: Width and Height for the glyph
+
+	sp: Source Pointer to data
+
+	span: Increment from line to line of data
+
+	Returns a pointer to the new glyph. Throws exception on failure to
+	allocate.
+*/
 fz_glyph *
 fz_new_glyph_from_1bpp_data(fz_context *ctx, int x, int y, int w, int h, unsigned char *sp, int span)
 {
